import { TypedEventEmitter, Libp2p } from "@libp2p/interface";
import { Logger } from "@libp2p/logger";
import { Node, Tree } from "@canvas-js/okra";
import type { Signature, Signer, Message, Awaitable } from "@canvas-js/interfaces";
import type { AbstractModelDB, ModelSchema, DatabaseUpgradeAPI, Config } from "@canvas-js/modeldb";
import { NetworkConfig, ServiceMap } from "@canvas-js/gossiplog/libp2p";
import { NetworkClient } from "@canvas-js/gossiplog/client";
import { SyncSnapshot } from "./interface.js";
import { AncestorIndex } from "./AncestorIndex.js";
import { MessageSet } from "./MessageSet.js";
import { MessageId } from "./MessageId.js";
import { MessageSource, SignedMessage } from "./SignedMessage.js";
export type GossipLogConsumer<Payload = unknown, Result = any> = (this: AbstractGossipLog<Payload, Result>, signedMessage: SignedMessage<Payload, Result>) => Awaitable<Result>;
export interface GossipLogInit<Payload = unknown, Result = any> {
    topic: string;
    apply: GossipLogConsumer<Payload, Result>;
    signer?: Signer<Payload>;
    /** validate that the IPLD `payload` is a `Payload` type */
    validatePayload?: (payload: unknown) => payload is Payload;
    verifySignature?: (signature: Signature, message: Message<Payload>) => Awaitable<void>;
    /** add extra tables to the local database for private use */
    schema?: ModelSchema;
    version?: Record<string, number>;
    upgrade?: (upgradeAPI: DatabaseUpgradeAPI, oldConfig: Config, oldVersion: Record<string, number>, newVersion: Record<string, number>) => Promise<boolean>;
    initialUpgradeSchema?: ModelSchema;
    initialUpgradeVersion?: Record<string, number>;
    clear?: boolean;
}
export type GossipLogEvents<Payload = unknown, Result = any> = {
    message: CustomEvent<SignedMessage<Payload>>;
    commit: CustomEvent<{
        root: Node;
        heads: string[];
    }>;
    sync: CustomEvent<{
        duration: number;
        messageCount: number;
        peer?: string;
    }>;
    connect: CustomEvent<{
        peer: string;
    }>;
    disconnect: CustomEvent<{
        peer: string;
    }>;
};
export type MessageRecord = {
    id: string;
    data: Uint8Array;
    hash: string;
    clock: number;
};
export type ReplayRecord = {
    timestamp: string;
    cursor: string | null;
};
export declare abstract class AbstractGossipLog<Payload = unknown, Result = any> extends TypedEventEmitter<GossipLogEvents<Payload, Result>> {
    #private;
    readonly db: AbstractModelDB;
    readonly tree: Tree;
    static namespace: string;
    static version: number;
    static schema: {
        $ancestors: {
            $primary: "key/clock";
            key: "bytes";
            clock: "integer";
            links: "bytes";
        };
        $messages: {
            id: "primary";
            data: "bytes";
            hash: "string";
            clock: "integer";
            $indexes: string[];
        };
        $heads: {
            id: "primary";
        };
        $replays: {
            timestamp: "primary";
            cursor: "string?";
            $indexes: string[];
        };
    };
    readonly topic: string;
    readonly signer: Signer<Payload>;
    readonly controller: AbortController;
    protected readonly log: Logger;
    protected readonly ancestorIndex: AncestorIndex;
    readonly validatePayload: (payload: unknown) => payload is Payload;
    readonly verifySignature: (signature: Signature, message: Message<Payload>) => Awaitable<void>;
    protected constructor(db: AbstractModelDB, tree: Tree, init: GossipLogInit<Payload, Result>);
    protected initialize(): Promise<void>;
    close(): Promise<void>;
    replay(): Promise<boolean>;
    connect(url: string, options?: {
        signal?: AbortSignal;
    }): Promise<NetworkClient<any>>;
    listen(port: number, options?: {
        signal?: AbortSignal;
    }): Promise<void>;
    startLibp2p(config: NetworkConfig): Promise<Libp2p<ServiceMap<Payload>>>;
    encode<T extends Payload = Payload>(signature: Signature, message: Message<T>, context?: {
        source?: MessageSource;
    }): SignedMessage<T, Result>;
    decode(value: Uint8Array, context?: {
        source?: MessageSource;
    }): SignedMessage<Payload, Result>;
    getHeads(): Promise<MessageSet>;
    getClock(): Promise<[clock: number, heads: string[]]>;
    has(id: string): Promise<boolean>;
    get(id: string): Promise<SignedMessage<Payload, Result> | null>;
    getMessages(range?: {
        lt?: string;
        lte?: string;
        gt?: string;
        gte?: string;
        reverse?: boolean;
        limit?: number;
    }): Promise<SignedMessage<Payload>[]>;
    iterate(range?: {
        lt?: string;
        lte?: string;
        gt?: string;
        gte?: string;
        reverse?: boolean;
        limit?: number;
    }): AsyncIterable<SignedMessage<Payload, Result>>;
    /**
     * Sign and append a new *unsigned* message to the end of the log.
     * The current concurrent heads of the local log are used as parents.
     */
    append<T extends Payload = Payload>(payload: T, { signer }?: {
        signer?: Signer<Payload>;
    }): Promise<SignedMessage<T, Result> & {
        result: Result;
    }>;
    /**
     * Insert an existing signed message into the log (ie received via HTTP API).
     * If any of the parents are not present, throw an error.
     */
    insert(signedMessage: SignedMessage<Payload, Result>): Promise<{
        id: string;
    }>;
    private apply;
    isAncestor(root: string | string[] | MessageId | MessageId[] | MessageSet, ancestor: string | MessageId): Promise<boolean>;
    /**
     * Sync with a remote source, applying and inserting all missing messages into the local log
     */
    sync(snapshot: SyncSnapshot, options?: {
        peer?: string;
    }): Promise<{
        complete: boolean;
        messageCount: number;
    }>;
    serve<T>(callback: (snapshot: SyncSnapshot) => Awaitable<T>): Promise<T>;
}
