import * as cbor from "@ipld/dag-cbor";
import { compare } from "uint8arrays";
import { assert } from "@canvas-js/utils";
import { MessageId, decodeId, getKey, encodeId, KEY_LENGTH } from "./MessageId.js";
import { MessageSet } from "./MessageSet.js";
import { decodeClock } from "./clock.js";
import { gossiplogTopicPattern } from "./utils.js";
export class SignedMessage {
    signature;
    message;
    value;
    parents;
    static decode(value, context = {}) {
        const [[codec, publicKey, signature], topic, clock, parentKeys, payload] = decodeMessageTuple(value);
        const parents = new MessageSet(parentKeys);
        const parentIds = [...parents].map((parent) => parent.id);
        return new SignedMessage({ codec, publicKey, signature }, { topic, clock, parents: parentIds, payload: payload }, value, parents, context);
    }
    static encode({ codec, publicKey, signature }, message, context = {}) {
        const { topic, clock, parents: parentIds, payload } = message;
        parentIds.sort();
        const parents = new MessageSet(parentIds);
        const parentKeys = [...parents].map((parent) => parent.key);
        const value = encodeMessageTuple([[codec, publicKey, signature], topic, clock, parentKeys, payload]);
        return new SignedMessage({ codec, publicKey, signature }, message, value, parents, context);
    }
    id;
    key;
    source;
    result;
    constructor(signature, message, value, parents, context) {
        this.signature = signature;
        this.message = message;
        this.value = value;
        this.parents = parents;
        this.key = getKey(message.clock, value);
        this.id = decodeId(this.key);
        this.source = context.source;
        this.result = context.result;
    }
}
function validateSignatureTuple(signatureTuple) {
    assert(Array.isArray(signatureTuple), "expected Array.isArray(signatureTuple)");
    assert(signatureTuple.length === 3, "expected signatureTuple.length === 5");
    const [codec, publicKey, signature] = signatureTuple;
    assert(typeof codec === "string", 'expected typeof codec === "string"');
    assert(typeof publicKey === "string", 'expected typeof publicKey === "string"');
    assert(signature instanceof Uint8Array, "expected signature instanceof Uint8Array");
}
function validateMessageTuple(messageTuple) {
    assert(Array.isArray(messageTuple), "expected Array.isArray(messageTuple)");
    assert(messageTuple.length === 5, "messageTuple.length === 5");
    const [signature, topic, clock, parents, payload] = messageTuple;
    validateSignatureTuple(signature);
    assert(typeof topic === "string", 'expected typeof topic === "string"');
    assert(gossiplogTopicPattern.test(topic), "invalid topic string");
    assert(typeof clock === "number", 'expected typeof clock === "number"');
    validateMessageParents(parents);
    assert((clock === 0 && parents.length === 0) || clock === getNextClock(parents), "expected clock === getNextClock(parents)");
}
function validateMessageParents(parents) {
    assert(Array.isArray(parents), "expected Array.isArray(parents)");
    for (const [i, parent] of parents.entries()) {
        assert(parent instanceof Uint8Array, "expected parent instanceof Uint8Array");
        assert(parent.length === KEY_LENGTH, "expected parent.length === KEY_LENGTH");
        if (i > 0) {
            assert(compare(parent, parents[i - 1]) === 1, "expected parents to be sorted lexicographically");
        }
    }
}
function encodeMessageTuple(messageTuple) {
    return cbor.encode(messageTuple);
}
function decodeMessageTuple(data) {
    const messageTuple = cbor.decode(data);
    validateMessageTuple(messageTuple);
    return messageTuple;
}
export function getNextClock(parents) {
    let max = 0;
    for (const parent of parents) {
        let clock;
        if (parent instanceof MessageId) {
            clock = parent.clock;
        }
        else if (typeof parent === "string") {
            clock = decodeClock(encodeId(parent))[0];
        }
        else {
            clock = decodeClock(parent)[0];
        }
        if (clock > max) {
            max = clock;
        }
    }
    return max + 1;
}
